unit fpprocesscapturepoller;

{$mode ObjFPC}

interface

uses
  Classes, SysUtils;

Type
  EProcessCapturePoller = class(Exception);

  { TProcessCapturePoller }
  TProcessStatus = (psRunning, // Process still running
                    psExited,  // Process has exited
                    psError    // Too many errors
                    );

  TOnGetProcessStatusEvent = Procedure (Sender : TObject; aProcessID : String; aOffset : NativeInt) of Object;
  TOnProcessDoneEvent = Procedure (Sender : TObject; aStatus : TProcessStatus; aExitCode : Integer) of Object;
  TOnProcessOutputEvent = Procedure (Sender : TObject; aOutput : String) of object;
  TOnStatusFailEvent = Procedure (Sender : TObject; aError : String) of object;

  { TCustomProcessCapturePoller }

  TCustomProcessCapturePoller = class(TComponent)
  private
    FBusy : Boolean;
    FLinebasedOutput: Boolean;
    FMaxCheckCount: Integer;
    FMaxFailCount: Integer;
    FOnGetProcessStatus: TOnGetProcessStatusEvent;
    FOnProcessDone: TOnProcessDoneEvent;
    FOnProcessOutput: TOnProcessOutputEvent;
    FOnStatusFail: TOnStatusFailEvent;
    FPollInterval: Integer;
    FProcessID: String;
    FCheckID : NativeInt;
    FCanceled : Boolean;
    FOutputOffset : NativeInt;
    FFailCount : Integer;
    FStatusCheckCount : Integer;
    procedure SetMaxCheckCount(AValue: Integer);
    procedure SetMaxFailCount(AValue: Integer);
    procedure SetProcessID(AValue: String);
  Protected
    procedure SetFailCount(aFailCount: Integer);
    procedure CancelCheck; virtual;
    procedure CheckBusy; virtual;
    procedure HandleOutput(aOffset : Integer; Const aOutput : String); virtual;
    Procedure ScheduleCheck; virtual;
    procedure ReportExit(aStatus : TProcessStatus; aExitCode : Integer); virtual;
    Procedure DoReportProgressFail(const aMessage : string); virtual;
    procedure DoStatusCheck; virtual; abstract;
    Procedure DoReportProgress(aStatus : TProcessStatus; aOutput : String; aExitCode : Integer; aOffSet : NativeInt); virtual;
  Public
    Constructor Create(aOwner : TComponent); override;
    destructor Destroy; override;
    Procedure Start;
    Procedure Cancel;

    Procedure ReportProgressFail(const aMessage : string);
    Property Canceled : Boolean Read FCanceled;
    Property FailCount : Integer Read FFailCount;
    Property StatusCheckCount : Integer Read FStatusCheckCount;
    Property OutputOffset : NativeInt Read FOutputOffset;
  Protected
    Property ProcessID : String Read FProcessID Write SetProcessID;
    Property OnProcessDone : TOnProcessDoneEvent Read FOnProcessDone Write FOnProcessDone;
    Property OnProcessOutput : TOnProcessOutputEvent Read FOnProcessOutput Write FOnProcessOutput;
    Property OnStatusFail : TOnStatusFailEvent Read FOnStatusFail Write FOnStatusFail;
    Property LinebasedOutput : Boolean Read FLinebasedOutput Write FLinebasedOutput default true;
    Property PollInterval : Integer Read FPollInterval Write FPollInterval default 500;
    Property MaxFailCount : Integer Read FMaxFailCount Write SetMaxFailCount default 1;
    Property MaxCheckCount : Integer Read FMaxCheckCount Write SetMaxCheckCount default 60;
  end;

  TProcessCapturePoller = Class(TCustomProcessCapturePoller)
  Protected
    procedure DoStatusCheck; override;
  Public
    Procedure ReportProgress(aStatus : TProcessStatus; aOutput : String; aExitCode : Integer; aOffSet : NativeInt); ;
  Published
    Property OnGetProcessStatus : TOnGetProcessStatusEvent Read FOnGetProcessStatus Write FOnGetProcessStatus;
    Property ProcessID;
    Property OnProcessDone;
    Property OnProcessOutput;
    Property OnStatusFail;
    Property LinebasedOutput;
    Property PollInterval;
    Property MaxFailCount;
    Property MaxCheckCount;
  end;
implementation

uses web;

Resourcestring
  SErrProcessBusy = 'Process is still busy';
  SErrNoProcessID = 'Can not start polling: Process ID not set';
  SErrErrorGettingStatus = 'Exception %s getting process status: %s';
  SErrMaxValueFailCount = 'Maximum value for MaxFailCount is %d, got %d';
  SErrMinValueCheckCount = 'Minimum value for MaxCheckCount is %d, got %d';

{ TCustomProcessCapturePoller }


procedure TCustomProcessCapturePoller.CheckBusy;

begin
  if FBusy then
    Raise EProcessCapturePoller.Create(SErrProcessBusy);
end;

procedure TCustomProcessCapturePoller.SetMaxCheckCount(AValue: Integer);
begin
  if FMaxCheckCount=AValue then Exit;
  if aValue<FMaxFailCount then
    Raise EProcessCapturePoller.CreateFmt(SErrMinValueCheckCount,[FMaxFailCount,AValue]);
  FMaxCheckCount:=AValue;
end;

procedure TCustomProcessCapturePoller.SetMaxFailCount(AValue: Integer);
begin
  if FMaxFailCount=AValue then Exit;
  if aValue>=FMaxCheckCount then
    Raise EProcessCapturePoller.CreateFmt(SErrMaxValueFailCount,[FMaxCheckCount,AValue]);
  FMaxFailCount:=AValue;
end;

procedure TCustomProcessCapturePoller.SetProcessID(AValue: String);
begin
  if FProcessID=AValue then Exit;
  CheckBusy;
  FProcessID:=AValue;
end;

procedure TCustomProcessCapturePoller.SetFailCount(aFailCount : Integer);
begin
  FFailCount:=aFailCount;
end;

procedure TProcessCapturePoller.DoStatusCheck;
begin
  try
    if Assigned(FOnGetProcessStatus) then
      FOnGetProcessStatus(Self,Self.FProcessID,Self.FOutputOffset)
  except
    On E : exception do
      ReportProgressFail(Format(SErrErrorGettingStatus,[E.ClassName,E.Message]));
  end;
end;

procedure TProcessCapturePoller.ReportProgress(aStatus: TProcessStatus; aOutput: String; aExitCode: Integer; aOffSet: NativeInt);
begin
  DoReportProgress(aStatus,aOutput,aExitCode,aOffset)
end;

constructor TCustomProcessCapturePoller.Create(aOwner: TComponent);
begin
  inherited Create(aOwner);
  FMaxCheckCount:=60;
  FPollInterval:=500;
  FMaxFailCount:=1;
  FLineBasedOutput:=True;
end;

destructor TCustomProcessCapturePoller.Destroy;
begin
  CancelCheck;
  inherited Destroy;
end;


procedure TCustomProcessCapturePoller.ScheduleCheck;

  Procedure CallStatusCheck;

  begin
    FCheckID:=0;
    DoStatusCheck;
  end;

begin
  FCheckID:=Window.setTimeout(@CallStatusCheck,500);
end;

procedure TCustomProcessCapturePoller.CancelCheck;

begin
 if FCheckID>0 then
   begin
   window.clearTimeout(FCheckID);
   FCheckID:=0;
   end;
end;

procedure TCustomProcessCapturePoller.ReportExit(aStatus: TProcessStatus; aExitCode: Integer);
begin
  CancelCheck;
  If Assigned(FOnProcessDone) then
    FOnProcessDone(Self,aStatus,aExitCode);
end;

procedure TCustomProcessCapturePoller.DoReportProgressFail(const aMessage: string);
begin
  Inc(FFailCount);
  If Assigned(FonStatusFail) then
    FOnStatusFail(Self,aMessage);
  if (FFailCount<FMaxFailCount) then
    ScheduleCheck
  else
    ReportExit(psError,-1)
end;

procedure TCustomProcessCapturePoller.Start;
begin
  CheckBusy;
  if (FProcessID='') then
    Raise EProcessCapturePoller.Create(SErrNoProcessID);
  FCanceled:=False;
  FFailCount:=0;
  FOutputOffset:=0;
  FStatusCheckCount:=FMaxCheckCount;
  ScheduleCheck;
end;

procedure TCustomProcessCapturePoller.Cancel;
begin
  CancelCheck;
  FCanceled:=True;
end;

procedure TCustomProcessCapturePoller.HandleOutput(aOffset: Integer; const aOutput: String);

var
  L : TStrings;
  aLine : String;

begin
  if aOffset>=0 then
    FOutputOffset:=aOffset;
  if not Assigned(FOnProcessOutput) then exit;
  if Not FLinebasedOutput then
    FOnProcessOutput(Self,aOutput)
  else
    begin
    L:=TStringList.Create;
    try
      L.Text:=aOUtput;
      For aLine in L do
        FOnProcessOutput(Self,aLine)
    finally
      L.Free;
    end;
    end;
end;

procedure TCustomProcessCapturePoller.DoReportProgress(aStatus: TProcessStatus; aOutput: String; aExitCode: Integer; aOffSet: NativeInt);
begin
  Dec(FStatusCheckCount);
  if FStatusCheckCount<=0 then
    DoReportProgressFail('Check timed out');
  case aStatus of
    psExited :
      begin
      HandleOutput(aOffset,aOutput);
      ReportExit(psExited,aExitCode);
      end;
    psRunning :
      begin
      HandleOutput(aOffset,aOutput);
      if (FCheckID=0) and not FCanceled then
        ScheduleCheck;
      end;
    psError:
      begin
      ReportProgressFail(aOutput);
      end;
  end
end;

procedure TCustomProcessCapturePoller.ReportProgressFail(const aMessage: string);
begin
  if FCanceled then
    exit;
  DoReportProgressFail(aMessage);
end;

end.

